/*
 * mmlist - doubly linked list library
 * Copyright (C) 1991-2001  Matthew Mondor
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Original author: Matthew Mondor, mmondor@linuxguru.net
 */



#ifndef MMLIST_H
#define MMLIST_H



/* HEADERS */

#ifndef MMTYPES_H
#include "mmtypes.h"
#endif




/* STRUCTURES */

typedef struct lnode
{
	struct lnode *ln_next;
	struct lnode *ln_prev;
}
lnode;

typedef struct list
{
	LONG l_lnodelen;
	ULONG l_lnodes;
	lnode *l_toplnode;
	lnode *l_bottomlnode;
	lnode *l_currentlnode;
}
list;




/* PROTOTYPES */

list *createlist(LONG);
list *freelist(list *);
lnode *appendnode(list *);
lnode *insertnode(list *);
lnode *deletenode(list *);
lnode *currentnode(list *);
lnode *nextnode(list *);
lnode *previousnode(list *);
lnode *topnode(list *);
lnode *bottomnode(list *);
ULONG totalnodes(list *);
VOID makecurrent(list *, lnode *);
lnode *scantonode(list *, ULONG);

/* These are useful for some optimizations, nodes can be linked and unlinked
 * in a very short time when mutexes are locked, then freed later on after
 * the mutex was released, same thing for allocation and linking; Since memory
 * management functions are usually what could slow it and cause a mutex to
 * be locked for long periods while preventing other threads to access the
 * shared list
 */
lnode *allocnode(list *);
lnode *freenode(lnode *);
VOID unlinknode(list *, lnode *);
lnode *linkappendnode(list *, lnode *);
lnode *linkinsertnode(list *, lnode *);

#endif

/*
	vim:ts=3:sw=3:cindent:
*/
