/*
 * mmtcpfwd v0.1.0 - Port forwarder superserver with DoS protection
 * Copyright (C) 2000, 2001  Matthew Mondor
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Original author: Matthew Mondor (mmondor@linuxguru.net)
 */




/* HEADERS */

#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>

#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <syslog.h>

#include <pthread.h>
#include <signal.h>
#include <sys/poll.h>
#include <time.h>

#include "mmfd.h"
#include "mmlist.h"




/* DEFINITIONS */

/* Useful when using pipe() generated fd pairs */
#define P_READ  0
#define P_WRITE 1

/* Configuration file columns definitions */
#define COL_PORT    0
#define COL_RESOLVE 1
#define COL_MAXIP   2
#define COL_MAXIPS  3
#define COL_MAXREJ  4
#define COL_TOHOST  5
#define COL_TOPORT  6
#define COL_TRANSP  7
#define COL_FTP     8
#define COL_TIMEOUT 9

/* Status return codes for a forwarder */
#define STATUS_UNKNOWN        0
#define CLIENT_CLOSED         1
#define SERVICE_CLOSED        2
#define CLIENT_READ_EOF       3
#define SERVICE_READ_EOF      4
#define CLIENT_WRITE_ERROR    5
#define SERVICE_WRITE_ERROR   6
#define SERVICE_CONNECT_ERROR 7
#define CREATE_SOCKET_ERROR   8
#define TRANSPROXY_BIND_ERROR 9
#define GIDUID_CHANGE_ERROR   10
#define PASV_BIND_ERROR       11
#define PASV_TIMEOUT_ERROR    12
#define CLOSED_SINCE_DENIED   13
#define BAD_IDENT_REQUEST     14
#define IDENT_OK              15
#define INPUT_TIMEOUT         16
#define UNAUTHORIZED_ADDRESS  17
#define PASV_ACCEPT_ERROR     18




/* STRUCTURES */

/* This structure is a node for a new IP entry that connected to us.
 */
typedef struct IPlistnode
{
	lnode node;

	LONG fromport;

	/* Client-specific info for each connection on this IP */
	struct sockaddr_in client;
	list *Threadlist;

	/* To control the connection attempts on a per-ip basis */
	LONG connections, refused;
	BOOL denied;

	/* Global information pointers */
	list *IPlist;
	pthread_mutex_t *IPlist_lock;
	BOOL resolve, transproxy, ftpproxy, identd;
	LONG uid, gid;
	BYTE **closed_status;
	BYTE *hostname;

	/* Host and port to forward connections to */
	BYTE *tohost;
	LONG toport;

	/* Inactivity timeout for connections */
	int timeout;
}
IPlistnode;

/* This structure is a node for each client for a single IP
 */
typedef struct Threadlistnode
{
	lnode node;

	struct IPlistnode *IPlnode;
	
	/* Information particular to this client */
	pthread_t thread;
	ULONG ID;
	
	/* Our network socket and pipe to receive signals from */
	int socket, tosocket;
	
	/* IP to connect to */
	struct sockaddr_in toserver;
}
Threadlistnode;

/* Information returned by start_rootserver() */
typedef struct rootdata
{
	int pid;
	int filedes[2];
}
rootdata;

/* Information to be passed to each fork()ed thread
 * (one per TCP port)
 */
typedef struct flistnode
{
	lnode node;

	rootdata *rdata;
	pthread_mutex_t *rootpipe_lock;
	LONG port, maxperip, maxips, maxrej, toport, mstimeout, uid, gid;
	BYTE tohost[80], *denymessage;
	BOOL resolve, transproxy, ftpproxy, identd;
	int pid;							  /* Set by start_fwdserver */
	FILE *closefh;
}
flistnode;

/* IP addresses to NOT be DENYed */
typedef struct niplistnode
{
	lnode node;
	struct in_addr inaddr;
}
niplistnode;

/* IP addresses that were DENYed if auto-restauration was used */
typedef struct diplistnode
{
	lnode node;
	time_t expire;
	struct in_addr inaddr;
}
diplistnode;




/* PROTOTYPES */

int main(int, char **);
static int start_fwdserver(flistnode *);
static VOID closeforwarders(IPlistnode *);
static BOOL addr_in_list(struct in_addr *);
static LONG readconfline(BYTE *, FILE *, LONG, LONG, BYTE **);
static LONG readconfline2(BYTE *, FILE *);
static rootdata *start_rootserver(FILE *, BYTE *, BYTE *, LONG);
VOID phandleclient(VOID *);
static int handleclient(int, IPlistnode *, Threadlistnode *);
static int ftpproxy(int, int, BYTE *, LONG, LONG, ULONG *, ULONG *,
						  Threadlistnode *);
static int handleident(int, IPlistnode *, Threadlistnode *);
static ULONG randname(BYTE *, ULONG);
static ULONG urand32(int, ULONG, ULONG);
static BOOL isvowel(BYTE);




/* GLOBAL VARIABLES */

static pthread_mutex_t rootpipe_lock;
static list *flist = NULL, *niplist = NULL;
static BYTE version[] = "\r\n+++ mmtcpfwd 0.1.0 (Matthew Mondor)\r\n";




/* MAIN PROGRAM */

int main(int argc, char *argv[])
{
	struct rootdata *rdata;
	struct in_addr inaddr;
	flistnode *flnode;
	niplistnode *niplnode;
	int ret = 0;
	LONG uid, gid, cols, denymins = 0;
	BYTE line[1024], *array[16], defcfg[] =
		"/etc/mmtcpfwd.conf", *denycommand, *undenycommand, *denymessage, *cfg =
		defcfg, *tmp;
	FILE *fh;
	BOOL ok;


	puts(version);

	openlog("mmtcpfwd", LOG_PID, LOG_AUTHPRIV);

	/* Make sure we were started by root */
	if(getuid()) {
		syslog(LOG_NOTICE, "Only root has rights to start daemon");
		exit(-1);
	}
	/* Initialize our global rootpipe_lock */
	pthread_mutex_init(&rootpipe_lock, NULL);


	/* Read our config file, binding ports and starting a process for each */
	if(argc == 2)
		cfg = argv[1];

	if(argc < 3) {
		if((fh = fopen(cfg, "r"))) {

			/* Get command to use to DENY an IP address from config file */
			if((cols = readconfline2(line, fh)) > 1) {
				if(!(denycommand = strdup(line))) {
					syslog(LOG_NOTICE, " * Out of memory error");
					fclose(fh);
					exit(-1);
				}
				syslog(LOG_NOTICE, "DENY-IP-Command: [%s]", denycommand);
			} else {
				syslog(LOG_NOTICE,
						 " * Wrong configuration file, first line should be the command line to DENY an IP address");
				fclose(fh);
				exit(-1);
			}

			/* Get command to use to unDENY an address that was DENYed */
			if((cols = readconfline2(line, fh)) > 1) {
				if(!(undenycommand = strdup(line))) {
					syslog(LOG_NOTICE, " * Out of memory error");
					fclose(fh);
					exit(-1);
				}
				syslog(LOG_NOTICE, "FREE-IP-Command: [%s]", undenycommand);
			} else {
				syslog(LOG_NOTICE,
						 " * Wrong configuration file, second line should be the command line to unDENY an IP address");
				fclose(fh);
				exit(-1);
			}

			/* Get message to send on DENY from config file */
			if((cols = readconfline2(line, fh)) > 1) {
				if(!(denymessage = strdup(line))) {
					syslog(LOG_NOTICE, " * Out of memory error");
					fclose(fh);
					exit(-1);
				}
				syslog(LOG_NOTICE, "DENY-IP-Message: [%s]", denymessage);
			} else {
				syslog(LOG_NOTICE,
						 " * Wrong configuration file, third line should be line sent for fake services before DENYing");
				fclose(fh);
				exit(-1);
			}

			/* Get the number of minutes an IP address should be DENYed for */
			if((cols = readconfline2(line, fh)) > 0) {
				denymins = atol(line);
			}
			if(!denymins) {
				syslog(LOG_NOTICE,
						 " * Wrong configuration file, fourth line should be the number of minutes an address should be DENYed for");
				fclose(fh);
				exit(-1);
			}
			/* Build our list of IP addresses to NOT DENY from conf file */
			ok = TRUE;
			if((niplist = createlist(sizeof(niplistnode)))) {
				while(ok) {
					if((cols = readconfline2(line, fh)) > 0) {
						if(*line == '.')
							break;
						if((inet_aton(line, &inaddr))) {
							/* Seems a valid IP address, add to our list */
							if((niplnode = (niplistnode *) appendnode(niplist))) {
								niplnode->inaddr.s_addr = inaddr.s_addr;
								if((tmp = inet_ntoa(inaddr)))
									syslog(LOG_NOTICE, "DENY-IP-Except.: [%s]", tmp);
							} else
								ok = FALSE;
						} else
							ok = FALSE;
					} else
						ok = FALSE;
				}
			} else
				ok = FALSE;

			if(!ok) {
				syslog(LOG_NOTICE,
						 " * Wrong configuration file, at the fifth line should be starting a list of IP addresses to NOT DENY, ending with a . line/entry");
				if(niplist)
					freelist(niplist);
				fclose(fh);
				exit(-1);
			}
			
			/* Start our rootserver process */
			syslog(LOG_NOTICE, "Starting RootServer process");
			if(!
				(rdata =
				 start_rootserver(fh, denycommand, undenycommand, denymins))) {
				syslog(LOG_NOTICE, " * Could not start RootServer process");
				fclose(fh);
				exit(-1);
			}
			
			/* Get uid and gid from config file */
			if((readconfline(line, fh, 2, 2, array)) == 2) {
				uid = atol(array[0]);
				gid = atol(array[1]);
			} else {
				syslog(LOG_NOTICE,
						 " * Wrong configuration file, sixth section should have uid gid column values");
				fclose(fh);
				exit(-1);
			}


			/* Now read all port forwarding entries, starting processes */
			if((flist = createlist(sizeof(flistnode)))) {
				while((cols = readconfline(line, fh, 4, 10, array)) != -1) {
					if(cols) {
						if(cols == 4) {
							if(!(atol(array[1]))) {
								if((flnode = (flistnode *) appendnode(flist))) {
									/* This is a fake service we deny IPs to */
									flnode->rdata = rdata;
									flnode->rootpipe_lock = &rootpipe_lock;
									flnode->port = atol(array[COL_PORT]);
									if(atol(array[COL_RESOLVE]))
										flnode->resolve = TRUE;
									else
										flnode->resolve = FALSE;
									flnode->maxperip = 0;
									flnode->maxips = atol(array[COL_MAXIPS]);
									flnode->maxrej = 0;
									flnode->tohost[0] = 0;
									flnode->denymessage = denymessage;
									flnode->toport = 0;
									flnode->mstimeout = 0;
									flnode->uid = uid;
									flnode->gid = gid;
									flnode->pid = 0;
									flnode->closefh = fh;

									if(!(start_fwdserver(flnode))) {
										syslog(LOG_NOTICE,
												 " * Error starting a fwd thread for port %ld",
												 flnode->port);
										ret = -1;
										break;
									}
								} else {
									syslog(LOG_NOTICE,
											 " * Error allocating memory for port %ld",
											 atol(array[COL_PORT]));
									ret = -1;
									break;
								}
							} else {
								syslog(LOG_NOTICE,
										 " * Error parsing config file (fake server should have 0/IP) for port %ld",
										 atol(array[COL_PORT]));
								ret = -1;
								break;
							}
						} else if(cols == 10) {
							if((flnode = (flistnode *) appendnode(flist))) {
								/* This is an actual service we forward */
								flnode->rdata = rdata;
								flnode->rootpipe_lock = &rootpipe_lock;
								flnode->port = atol(array[COL_PORT]);
								if(atol(array[COL_RESOLVE]))
									flnode->resolve = TRUE;
								else
									flnode->resolve = FALSE;
								flnode->maxperip = atol(array[COL_MAXIP]);
								flnode->maxips = atol(array[COL_MAXIPS]);
								flnode->maxrej = atol(array[COL_MAXREJ]);
								strncpy(flnode->tohost, array[COL_TOHOST], 78);
								flnode->denymessage = denymessage;
								flnode->toport = atol(array[COL_TOPORT]);
								if(atol(array[COL_TRANSP]))
									flnode->transproxy = TRUE;
								else
									flnode->transproxy = FALSE;
								if(atol(array[COL_FTP]))
									flnode->ftpproxy = TRUE;
								else
									flnode->ftpproxy = FALSE;
								if(flnode->port == 113 && !(atoi(flnode->tohost))
									&& flnode->toport == 0)
									flnode->identd = TRUE;
								else
									flnode->identd = FALSE;
								flnode->mstimeout = atol(array[COL_TIMEOUT]);
								flnode->uid = uid;
								flnode->gid = gid;
								flnode->pid = 0;
								flnode->closefh = fh;

								if(!(start_fwdserver(flnode))) {
									syslog(LOG_NOTICE,
											 " * Error starting a fwd daemon process on port %ld",
											 atol(array[COL_PORT]));
									ret = -1;
									break;
								}
							} else {
								syslog(LOG_NOTICE,
										 " * Error allocating memory for port %ld",
										 atol(array[COL_PORT]));
								ret = -1;
								break;
							}
						} else {
							syslog(LOG_NOTICE,
									 " * Error in configuration file entry");
							ret = -1;
							break;
						}
					}
				}
			}
			if(rdata)
				free(rdata);
			if(ret == -1) {
				if(fh)
					fclose(fh);
				if(flist)
					freelist(flist);
				exit(ret);
			}
		} else {
			syslog(LOG_NOTICE, " * Could not open config file (%s)", cfg);
			exit(-1);
		}
	} else {
		puts("USAGE: mmtcpfwd [<configfile>]\r\n");
		exit(-1);
	}

	exit(0);
}


/* This is the actual main program which handles forwarding for each port. */
int start_fwdserver(flistnode * fnode)
{
	struct sockaddr addr;
	struct sockaddr_in server, *saddr = (struct sockaddr_in *)&addr;

	pthread_mutex_t IPlist_lock;
	list *IPlist;

	socklen_t addrl;
	IPlistnode *IPlnode;
	Threadlistnode *Threadlnode;

	int sock, msgsock = -1, ret, pid, fd;
	pthread_attr_t threadattr;	  /* Shared pthreads Attributes */
	BYTE *tmp;
	BOOL found, ok, deny;
	ULONG ID = 0;


	/* We keep this array and share it with all the pthreads
	  	handling connections
	 */
	BYTE *closed_status[] = {
		"STATUS_UNKNOWN",
		"CLIENT_CLOSED",
		"SERVICE_CLOSED",
		"CLIENT_READ_EOF",
		"SERVICE_READ_EOF",
		"CLIENT_WRITE_ERROR",
		"SERVICE_WRITE_ERROR",
		"SERVICE_CONNECT_ERROR",
		"CREATE_SOCKET_ERROR",
		"TRANSPROXY_BIND_ERROR",
		"GIDUID_CHANGE_ERROR",
		"PASV_BIND_ERROR",
		"PASV_TIMEOUT_ERROR",
		"CLOSED_SINCE_DENIED",	  /* Not actually used at present time */
		"BAD_IDENT_REQUEST",
		"IDENT_OK",
		"INPUT_TIMEOUT",
		"UNAUTHORIZED_ADDRESS",
		"PASV_ACCEPT_ERROR"
	};


	if(!(pid = fork())) {

		/* Child process, detach... */
		setsid();
		chdir("/");
		if((fd = open("/dev/null", O_RDWR)) != -1) {
			dup2(fd, 0);
			dup2(fd, 1);
			dup2(fd, 2);
			if(fd > 2)
				close(fd);
		}
		signal(SIGTTOU, SIG_IGN);
		signal(SIGTTIN, SIG_IGN);
		signal(SIGTSTP, SIG_IGN);
		umask(0);

		/* Close file handles we shouldn't have access to */
		close(fnode->rdata->filedes[P_READ]);
		fclose(fnode->closefh);

		/* Make our own copy of the denymessage string */
		if((tmp = strdup(fnode->denymessage)))
			fnode->denymessage = tmp;

		openlog("mmtcpfwd", LOG_PID, LOG_AUTHPRIV);


		if(fnode->maxperip) {
			if(fnode->identd)
				syslog(LOG_NOTICE, "Starting mmidentd on port %d", fnode->port);
			else if(fnode->ftpproxy)
				syslog(LOG_NOTICE, "Starting ftp proxy on port %d", fnode->port);
			else
				syslog(LOG_NOTICE, "Starting forwarder on port %d", fnode->port);
		} else
			syslog(LOG_NOTICE, "Starting fake service on port %d", fnode->port);


		/* Prepare pthread attributes, as well as clientslist lock.
		 * we want our threads to be detached, non-joinable, as we want
		 * them to auto-release their ressources when exiting
		 */
		pthread_attr_init(&threadattr);
		pthread_attr_setdetachstate(&threadattr, PTHREAD_CREATE_DETACHED);
		pthread_mutex_init(&IPlist_lock, NULL);


		/* Create our IP addresses list for this port */
		if((IPlist = createlist(sizeof(IPlistnode)))) {

			/* This section binds a port, listens for connections to it, and
			 * dispaches each client connecting to it to a multi-tasking client
			 * server thread
			 */
			ret = 0;
			if((sock = socket(AF_INET, SOCK_STREAM, 0)) > -1) {

				memset(&server, 0, sizeof(server));
				server.sin_family = AF_INET;
				server.sin_addr.s_addr = INADDR_ANY;
				server.sin_port = htons(fnode->port);

				if(!
					(bind
					 (sock, (struct sockaddr *)(void *)&server, sizeof(server)))) {

					/* Become non-root user and group (usually nobody/nogroup) */
					if((!(setegid(fnode->gid))) && (!(seteuid(fnode->uid)))) {

						if(!(listen(sock, fnode->maxips * fnode->maxperip))) {

							/* Wait for client connections and handle them */
							while(TRUE) {

								addrl = sizeof(struct sockaddr);
								if((msgsock = accept(sock, &addr, &addrl)) != -1) {

									/* Lock the list so that no other process
									 * or thread can access it
									 */
									pthread_mutex_lock(&IPlist_lock);

									/* Try to locate the IP address of the new
									 * connection in our IPlist. This uses
									 * the linked list structures directly for
									 * maximum speed
									 */
									found = FALSE;
									IPlnode = (IPlistnode *) IPlist->l_toplnode;
									while(IPlnode) {
										if(IPlnode->client.sin_addr.s_addr ==
											saddr->sin_addr.s_addr) {
											found = TRUE;
											break;
										}
										IPlnode = (IPlistnode *) IPlnode->node.ln_next;
									}
									ok = FALSE;

									/* At this point we know if the current
									 * client's address already exists in our
									 * list (if there are connections from it)
									 */

									/* If maxperip==0, we simply immediately
									 * close connection and DENY this IP address,
									 * Then continue
									 */
									if(!fnode->maxperip)
										deny = TRUE;
									else {

										/* Normal forwarder operation */
										deny = FALSE;

										if(found) {

											/* At least one connection is
											 * forwarded for this IP Address.
											 * Make sure that this it was not
											 * DENYed yet.
											 */
											if(!IPlnode->denied) {
												if(IPlnode->connections < fnode->maxperip)
													/* This IP is still allowed
													 * connections, allow starting
													 * a forwarder for it
													 */
													ok = TRUE;
												else {
													/* The client attempts to
													 * exceed the allowed
													 * simultanious connections
													 * for this IP address
													 */
													IPlnode->refused++;
													if(IPlnode->refused > fnode->maxrej)
														/* Client has exceeded
														 * the number of allowed
														 * failed connections
														 * attempts for this IP
														 * address. He is DoSing.
														 * Log it, and DENY that
														 * IP address.
														 */
														deny = TRUE;
												}
											}

										} else {

											/* This IP not found in IPlist,
											 * If maximum total IP isn't reached,
											 * add a new node for it. 
											 */
											if(IPlist->l_lnodes < fnode->maxips) {
												if((IPlnode = (IPlistnode *)
													 appendnode(IPlist))) {
													/* We set IPlnode to be
													 * shared with all clients
													 * for this IP, allocate the
													 * list for clients to use
													 * it, and allow starting a
													 * forwarder
													 */
													if((IPlnode->Threadlist =
														 createlist(sizeof
																		(Threadlistnode)))) {
														IPlnode->client = *saddr;
														IPlnode->IPlist = IPlist;
														IPlnode->IPlist_lock = &IPlist_lock;
														IPlnode->closed_status =
															closed_status;
														IPlnode->tohost = fnode->tohost;
														IPlnode->fromport = fnode->port;
														IPlnode->toport = fnode->toport;
														IPlnode->timeout = fnode->mstimeout;
														IPlnode->resolve = fnode->resolve;
														IPlnode->transproxy =
															fnode->transproxy;
														IPlnode->ftpproxy = fnode->ftpproxy;
														IPlnode->identd = fnode->identd;
														IPlnode->uid = fnode->uid;
														IPlnode->gid = fnode->gid;
														ok = TRUE;
													} else
														/* Out of memory */
														syslog(LOG_NOTICE,
																 "DEBUG: Out of memory creating IPlnode Threadlist for port %d",
																 fnode->port);
												} else
													/* Out of memory */
													syslog(LOG_NOTICE,
															 "DEBUG: Out of memory appending new IPlistnode for port %d",
															 fnode->port);
											} else
												/* Max allowed IP addresses reached */
												syslog(LOG_NOTICE,
														 "DEBUG: Maximum IPs reached for port %d",
														 fnode->port);
										}

									}


									if(deny) {

										/* We are supposed to DENY this address */
										ok = FALSE;

										if(!fnode->maxperip)
											/* Send warning message */
											fdprintf(msgsock, "\r\n%s\r\n\r\n",
														fnode->denymessage);

										/* Close socket immediately */
										shutdown(msgsock, 2);
										close(msgsock);
										msgsock = -1;

										/* Now DENY this address if not in our
										 * not to deny list
										 */
										if(!addr_in_list(&saddr->sin_addr)) {

											if(found) {
												if(!IPlnode->denied) {
													IPlnode->denied = TRUE;
													/* Tell client threads to quit */
													closeforwarders(IPlnode);
													/* From now on this IPlnode is
													 * no longer allocated, we must
													 * not access it anymore
													 */
												} else
													deny = FALSE;
											}

											if(deny) {

												/* Address was not yet denied,
												 * tell our rootserver to do it
												 */
												if((tmp = inet_ntoa(saddr->sin_addr))) {
													pthread_mutex_lock(fnode->
																			 rootpipe_lock);
													if((fdprintf(fnode->rdata->
																	 filedes[P_WRITE],
																	 "%s:%d\n", tmp,
																	 fnode->port)) == -1)
														syslog(LOG_NOTICE,
																 "DEBUG: Error writing to rootserver pipe for port %d",
																 fnode->port);
													pthread_mutex_unlock(fnode->
																				rootpipe_lock);
												} else
													syslog(LOG_NOTICE,
															 "DEBUG: inet_ntoa() failed to DENY address for port %d",
															 fnode->port);

											}

										}

									}

									if(ok) {

										/* We were allowed to start a forwarder,
										 * Add a node to our Threadlist in this
										 * IPlistnode, and start thread to serve
										 * the client. Here again, allocation
										 * is performed within the shared list
										 * locked mutex.
										 */
										if((Threadlnode = (Threadlistnode *)
											 appendnode(IPlnode->Threadlist))) {
											Threadlnode->socket = msgsock;
											Threadlnode->tosocket = -1;
											ID++;
											Threadlnode->ID = ID;
											Threadlnode->IPlnode = IPlnode;
											/* Finally, start the forwarder
											 * thread.
											 */
											if(!
												(pthread_create
												 (&(Threadlnode->thread),
												  &threadattr,
												  (VOID *) phandleclient,
												  (VOID *) Threadlnode)))
												/* Increase our connections
												 * counter for this IPlistnode
												 */
												IPlnode->connections++;
											else
												/* Cannot start thread */
												syslog(LOG_NOTICE,
														 "DEBUG: Error starting thread for port %d",
														 fnode->port);
										} else
											/* Out of memory */
											syslog(LOG_NOTICE,
													 "DEBUG: Out of memory allocating Threadlnode for port %d",
													 fnode->port);

									} else {

										/* We just close immediately this
										  	connection attempt
										 */
										if(msgsock != -1) {
											shutdown(msgsock, 2);
											close(msgsock);
											msgsock = -1;
										}

									}

									/* Unlock the mutex since we are done with
									 * it.
									 */
									pthread_mutex_unlock(&IPlist_lock);

								} else
									syslog(LOG_NOTICE,
											 " * Cannot accept connection for port %d",
											 fnode->port);

								/* Continue main loop here */
							}

						} else {
							syslog(LOG_NOTICE,
									 "Cannot listen to socket for port %d",
									 fnode->port);
							ret = -1;
						}
					} else {
						syslog(LOG_NOTICE,
								 "Cannot change uid and gid for port %d!",
								 fnode->port);
						ret = -1;
					}

				} else {
					syslog(LOG_NOTICE, "Cannot bind socket for port %d",
							 fnode->port);
					ret = -1;
				}

				shutdown(sock, 2);
				close(sock);

			} else {
				syslog(LOG_NOTICE, "Cannot create socket for port %d",
						 fnode->port);
				ret = -1;
			}

			IPlist = freelist(IPlist);
		} else {
			syslog(LOG_NOTICE, "Out of memory for port %d", fnode->port);
			ret = -1;
		}

		exit(ret);

	} else if(pid > 0) {

		/* Parent */
		fnode->pid = pid;
		return (pid);

	}

	return (0);
}


/* Signal all forwarders of a list to cleanly exit. This is to ensure that all
 * forwarders cleanly quit before we DENY an IP address, to not have partial
 * TCP connections pending. We also need to free the ressources that were
 * allocated by the main process for them. The shared IPlist mutex is already
 * locked for us my the main forwarder function.
 */
VOID closeforwarders(IPlistnode * IPlnode)
{
	list *Threadlist, *IPlist;
	Threadlistnode *Threadlnode;

	if(IPlnode) {
		if((Threadlist = IPlnode->Threadlist)) {

			Threadlnode = (Threadlistnode *) Threadlist->l_toplnode;
			while(Threadlnode) {

				pthread_cancel(Threadlnode->thread);

				/* Cleanup ressources we associated to forwarder */
				if(Threadlnode->socket != -1) {
					shutdown(Threadlnode->socket, 2);
					close(Threadlnode->socket);
					Threadlnode->socket = -1;
				}
				if(Threadlnode->tosocket != -1) {
					shutdown(Threadlnode->tosocket, 2);
					close(Threadlnode->tosocket);
					Threadlnode->tosocket = -1;
				}

				Threadlnode = (Threadlistnode *) Threadlnode->node.ln_next;
			}

			IPlnode->Threadlist = NULL;
			freelist(Threadlist);
		}

		if(IPlnode->hostname) {
			free(IPlnode->hostname);
			IPlnode->hostname = 0;
		}
		if((IPlist = IPlnode->IPlist)) {
			makecurrent(IPlist, (lnode *) IPlnode);
			deletenode(IPlist);
		}
	}
}


BOOL addr_in_list(struct in_addr *inaddr)
{
	niplistnode *niplnode;
	BOOL found = FALSE;

	niplnode = (niplistnode *) niplist->l_toplnode;
	while(niplnode) {
		if(niplnode->inaddr.s_addr == inaddr->s_addr) {
			found = TRUE;
			break;
		}
		niplnode = (niplistnode *) niplnode->node.ln_next;
	}

	return (found);
}


/* This function reads a valid (non-blank, non-comment) line from fh,
 * and parses the columns of that line, filling array with the pointers
 * to each trim string. It returns 0 if the number of columns of that line
 * is out of mincols,maxcols bounds, the number of columns of the line on
 * success, or -1 on EOF.
 */
LONG readconfline(BYTE * str, FILE * fh, LONG mincols, LONG maxcols,
						BYTE ** array)
{
	BYTE *ptr, *ptr2;
	LONG len, col;

	maxcols++;

	while(fgets(str, 1023, fh)) {
		len = strlen(str);

		if(str[len - 1] == '\n') {
			len--;
			str[len] = 0;
		}
		if(str[len - 1] == '\r') {
			len--;
			str[len] = 0;
		}

		if(len) {
			/* Comment or empty line? */
			ptr = str;
			while(*ptr && (*ptr == ' ' || *ptr == '\t'))
				ptr++;
			if(!*ptr || *ptr == '#' || *ptr == ';')
				continue;

			/* No, so now format it's columns */
			col = 0;
			ptr = str;

			while((*ptr) && (col < maxcols)) {

				/* Find first printable char */
				while(*ptr && (*ptr == ' ' || *ptr == '\t'))
					ptr++;

				if(*ptr) {
					ptr2 = ptr;

					/* Find next unprintable char */
					while(*ptr && *ptr != ' ' && *ptr != '\t')
						ptr++;

					if(ptr != ptr2) {
						if(*ptr) {
							*ptr = 0;
							ptr++;
						}
						array[col] = ptr2;
						col++;
					} else
						break;

				} else
					break;

			}

			if((col < maxcols) && (col >= mincols))
				return (col);

			return (0);

		}
	}

	return (-1);
}


LONG readconfline2(BYTE * str, FILE * fh)
{
	LONG len;
	BYTE *ptr;

	while(fgets(str, 1023, fh)) {
		len = strlen(str);

		if(str[len - 1] == '\n') {
			len--;
			str[len] = 0;
		}
		if(str[len - 1] == '\r') {
			len--;
			str[len] = 0;
		}

		if(len) {
			/* Comment or empty line? */
			ptr = str;
			while(*ptr && (*ptr == ' ' || *ptr == '\t'))
				ptr++;
			if(!*ptr || *ptr == '#' || *ptr == ';')
				continue;

			return (len);
		}
	}

	return (-1);
}


/* This function starts a background process, the root server. This is
 * done before using setuid() to become user nobody. Returns 0/NULL on error,
 * or returns the fd to send rootserver commands. It seems that pthreads
 * could not be used for this task, I thus had to fork() once again.
 * We also take care to undeny IP addresses which have been denied long enough
 */
rootdata *start_rootserver(FILE * fh, BYTE * odenycommand,
									BYTE * oundenycommand, LONG mins)
{
	list *diplist = NULL;
	diplistnode *diplnode;
	rootdata *rdata = NULL;
	int pid, fd;
	LONG len;
	BYTE buffer[1024], ipaddr[20], *ptr, *tmp, *denycommand =
		NULL, *undenycommand = NULL;
	time_t curtime;

	/* Allocate structure */
	if((rdata = (rootdata *) calloc(1, sizeof(rootdata)))) {

		/* Create our pipe (filedes[0]=read, filedes[1]=write) */
		if(!(pipe(rdata->filedes))) {
			/* Pipe successfully created, start our rootserver process */
			if(!(pid = fork())) {

				/* Child process, actual root process. Detach... */
				setsid();
				chdir("/");
				if((fd = open("/dev/null", O_RDWR)) != -1) {
					dup2(fd, 0);
					dup2(fd, 1);
					dup2(fd, 2);
					if(fd > 2)
						close(fd);
				}
				signal(SIGTTOU, SIG_IGN);
				signal(SIGTTIN, SIG_IGN);
				signal(SIGTSTP, SIG_IGN);
				umask(0);

				/* Close filehandles we shouldn't have access to */
				close(rdata->filedes[P_WRITE]);
				fclose(fh);
				rdata->filedes[P_WRITE] = -1;
				fh = NULL;

				openlog("mmtcpfwd", LOG_PID, LOG_AUTHPRIV);

				/* Make our own copy of the (un)denycommand strings */
				denycommand = strdup(odenycommand);
				undenycommand = strdup(oundenycommand);

				if(denycommand && undenycommand
					&& (diplist = createlist(sizeof(diplistnode)))) {

					while(1) {

						struct in_addr inaddr;
						int tmpret;

						/* Here we wait for a line to be read for at most
						  	60 seconds
						 */
						if((len =
							 fdgets(buffer, 1000, rdata->filedes[P_READ],
									  60000)) != -2) {

							/* The string we received should be in the format:
							 * n.n.n.n:p, make sure it is, isolate IP and port
							 */
							ptr = buffer;
							while(*ptr && *ptr != ':')
								ptr++;
							if(*ptr == ':') {
								*ptr = 0;
								ptr++;

								/* Make sure this is a valid IP address */
								if(inet_aton(buffer, &inaddr)) {

									strncpy(ipaddr, buffer, 19);
									syslog(LOG_NOTICE,
											 "RootServer: DoS from [%s] on port %s (DENYING)",
											 ipaddr, ptr);
									snprintf(buffer, 1023, denycommand, ipaddr);

									tmpret = system(buffer);
									if(tmpret != -1 && tmpret != 127) {

										if(mins > 0) {
											/* Add new timeout event to undeny it */
											if((diplnode =
												 (diplistnode *) appendnode(diplist))) {
												diplnode->expire = time(NULL);
												diplnode->expire += (mins * 60);
												diplnode->inaddr = inaddr;
											} else
												syslog(LOG_NOTICE,
														 "RootServer: Out of memory setting up undeny event for port %s",
														 ptr);
										}

									} else
										syslog(LOG_NOTICE,
												 "RootServer: Error executing deny command for port %s",
												 ptr);

								} else
									syslog(LOG_NOTICE,
											 "RootServer: Received invalid address for port %s",
											 ptr);

							}

						} else {
							syslog(LOG_NOTICE,
									 "RootServer: Error reading from pipe");
							break;
						}

						/* Process undeny events if any */
						if(mins > 0) {

							curtime = time(NULL);

							diplnode = (diplistnode *) diplist->l_toplnode;
							while(diplnode) {

								if(diplnode->expire <= curtime) {
									/* This entry has expired, undeny this IP
									 * address and free the entry
									 */
									if((tmp = inet_ntoa(diplnode->inaddr))) {
										syslog(LOG_NOTICE,
												 "RootServer: UNDENYING [%s]", tmp);
										snprintf(buffer, 1023, undenycommand, tmp);
										tmpret = system(buffer);
										if(tmpret == -1 || tmpret == 127)
											syslog(LOG_NOTICE,
													 "RootServer: Could not execute undeny command");
									} else
										syslog(LOG_NOTICE,
												 "RootServer: Could not convert address to ASCII, not undenying");
									makecurrent(diplist, (lnode *) diplnode);
									diplnode = (diplistnode *) diplnode->node.ln_next;
									deletenode(diplist);
									continue;
								}

								diplnode = (diplistnode *) diplnode->node.ln_next;
							}

						}

					}					  /* while */

					syslog(LOG_NOTICE, "RootServer: Quitting");

					freelist(diplist);
				} else
					syslog(LOG_NOTICE,
							 "RootServer: Out of memory, could not start");

				if(denycommand)
					free(denycommand);
				if(undenycommand)
					free(undenycommand);

				closelog();
				exit(0);

			}
			if(pid > 0) {

				/* Child started successfully, this is parent */
				rdata->pid = pid;
				return (rdata);

			}
			/* Error */
		}

		if(rdata)
			free(rdata);

	}

	return (NULL);
}




/* SERVER (THREAD) */

/* This is started by pthread_create, it calls handleclients(), then calls
 * pcleanupclient() to release the ressources that were allocated to us
 * by our parent process
 */
VOID phandleclient(VOID * args)
{
	int ret = 0, socket, oldtype;
	Threadlistnode *Threadlnode = (Threadlistnode *) args;
	IPlistnode *IPlnode = Threadlnode->IPlnode;
	pthread_mutex_t *IPlist_lock = IPlnode->IPlist_lock;
	list *IPlist;
	struct in_addr *ip;
	struct hostent *hoste;
	BYTE *tmp, *tmp2;;

	/* Enable cancelation requests from pthread_cancel() */
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	openlog("mmtcpfwd", LOG_PID, LOG_AUTHPRIV);

	socket = Threadlnode->socket;

	/* Here we are using a method which only requires to lock the mutex
	 * once when writing to the shared data
	 */

	/* We want to resolve the client IP's hostname, but not if it already been
	 * done by another client on this same IP. So do it if it was not,
	 * only if we want to resolve them.
	 */
	tmp = NULL;
	if((IPlnode->resolve) && (!IPlnode->hostname)) {
		if((hoste =
			 gethostbyaddr((char *)&(IPlnode->client.sin_addr),
								sizeof(struct in_addr), AF_INET))) {
			if(!(tmp = strdup(hoste->h_name)))
				syslog(LOG_NOTICE, "DEBUG: out of memory for port %d",
						 IPlnode->fromport);
		} else {
			/* Could not resolve hostname, reply why in logfile */
			switch (h_errno) {
				case HOST_NOT_FOUND:
					tmp2 = "HOST_NOT_FOUND";
					break;
				case NO_ADDRESS:
					tmp2 = "NO_ADDRESS";
					break;
					/*
						case NO_DATA:
						   tmp2="NO_DATA";
						   break;
					*/
				case NO_RECOVERY:
					tmp2 = "NO_RECOVERY";
					break;
				case TRY_AGAIN:
					tmp2 = "TRY_AGAIN";
					break;
				default:
					tmp2 = "UNKNOWN_ERROR";
			}
			syslog(LOG_NOTICE,
					 "DEBUG: gethostbyaddr failed with %08X for port %d (%s)",
					 IPlnode->client.sin_addr, IPlnode->fromport, tmp2);

			if(!(tmp = strdup("unknown")))
				syslog(LOG_NOTICE, "DEBUG: out of memory for port %d",
						 IPlnode->fromport);
		}
	}

	/* Here we resolve the IP of the hostname we must connect to.
	 * This is done everytime, since some hostnames return changing
	 * IPs. This will be quick if an IP was supplied instead of a hostname.
	 */
	ip = NULL;
	if((hoste = gethostbyname(IPlnode->tohost)))
		ip = (struct in_addr *)hoste->h_addr_list[0];
	else
		syslog(LOG_NOTICE,
				 "DEBUG: Cannot resolve hostname to forward port %d",
				 IPlnode->fromport);

	if(tmp || ip) {
		/* Here we lock the mutex once and perform needed writes to the shared
		 * list. This prevents the mutex to be locked and unlocked too often.
		 * Make sure to prevent any possible cancellation request
		 * from breaking the thread while the mutex is locked.
		 * We thus set the cancel type to DEFERRED temporarily,
		 * and set a cleanup handler to unlock it if it was locked.
		 */
		pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &oldtype);
		pthread_cleanup_push((VOID *) pthread_mutex_unlock,
									(VOID *) IPlist_lock);
		pthread_mutex_lock(IPlist_lock);

		/* Perform writes */

		if(tmp)
			IPlnode->hostname = tmp;

		if(ip) {
			/* Setup our toserver structure */
			memset(&(Threadlnode->toserver), 0, sizeof(struct sockaddr_in));
			Threadlnode->toserver.sin_family = AF_INET;
			Threadlnode->toserver.sin_addr.s_addr = ip->s_addr;
			Threadlnode->toserver.sin_port = htons((UWORD) IPlnode->toport);
		}

		/* Unlock mutex, get rid of cleanup handler and restore normal cancel
		 * type
		 */
		pthread_mutex_unlock(IPlist_lock);
		pthread_cleanup_pop(0);
		pthread_setcanceltype(oldtype, NULL);
	}


	if(ip) {
		/* Handle the client. */
		if(IPlnode->identd)
			ret = handleident(socket, IPlnode, Threadlnode);
		else
			ret = handleclient(socket, IPlnode, Threadlnode);
	}
	/* Cleanup; close connection with client, etc */
	IPlist = Threadlnode->IPlnode->IPlist;

	/* Lock our shared list mutex */
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &oldtype);
	pthread_cleanup_push((VOID *) pthread_mutex_unlock, (VOID *) IPlist_lock);
	pthread_mutex_lock(IPlist_lock);

	/* Make sure that both our service and client sockets are closed */
	if(Threadlnode->socket != -1) {
		shutdown(Threadlnode->socket, 2);
		close(Threadlnode->socket);
		Threadlnode->socket = -1;
	}
	if(Threadlnode->tosocket != -1) {
		shutdown(Threadlnode->tosocket, 2);
		close(Threadlnode->tosocket);
		Threadlnode->tosocket = -1;
	}
	/* Delete our Threadlnode */
	makecurrent(IPlnode->Threadlist, (lnode *) Threadlnode);
	deletenode(IPlnode->Threadlist);

	/* Decrease connections counter for this IP */
	if(IPlnode->connections)
		IPlnode->connections--;

	if(!(IPlnode->connections)) {
		/* No more connections for this IP address, free it's Threadlist, it's
		 * possibly resolved IP hostname, and delete this IPlnode
		 */
		if(IPlnode->Threadlist)
			freelist(IPlnode->Threadlist);
		if(IPlnode->hostname)
			free(IPlnode->hostname);
		makecurrent(IPlist, (lnode *) IPlnode);
		deletenode(IPlist);
	}
	/* Unlock shared list mutex and free cancel handler */
	pthread_mutex_unlock(IPlist_lock);
	pthread_cleanup_pop(0);
	pthread_setcanceltype(oldtype, NULL);

	/* Exit thread */
	closelog();
	pthread_exit(&ret);
}


/* This function handles a client - connects to the wanted service and
 * forwards packets in and out, doing stats and loging. It is called from
 * phandleclient() function/thread. It also calls ftpproxy() if required.
 */
int handleclient(int fd, IPlistnode * IPlnode, Threadlistnode * Threadlnode)
{
	struct sockaddr_in addr_out;
	BYTE buffer[8192], ipaddr[20], toipaddr[20];
	int tofd = -1, selr, timeout;
	BYTE *tmp;
	LONG len, cnt;
	BOOL ok;

	ULONG bytes_in, bytes_out, time_total;
	time_t time_start, time_end;
	int status;


	/* Obtain IP address of client */
	if((tmp = inet_ntoa(IPlnode->client.sin_addr)))
		strncpy(ipaddr, tmp, 19);
	else
		strncpy(ipaddr, "0.0.0.0", 8);

	/* Obtain IP address of server */
	if((tmp = inet_ntoa(Threadlnode->toserver.sin_addr)))
		strncpy(toipaddr, tmp, 19);
	else
		strncpy(toipaddr, "0.0.0.0", 8);

	if(IPlnode->hostname)
		/* Log user's IP and hostname */
		syslog(LOG_NOTICE, "Connect: ID=0x%04X%08X [%s:%d] - (%s) -> [%s:%d]",
				 IPlnode->fromport, Threadlnode->ID, ipaddr, IPlnode->fromport,
				 IPlnode->hostname, toipaddr, IPlnode->toport);
	else
		/* Log user's IP */
		syslog(LOG_NOTICE, "Connect: ID=0x%04X%08X [%s:%d] -> [%s:%d]",
				 IPlnode->fromport, Threadlnode->ID, ipaddr, IPlnode->fromport,
				 toipaddr, IPlnode->toport);

	bytes_in = bytes_out = 0;
	time_start = time(NULL);
	status = STATUS_UNKNOWN;

	ok = FALSE;
	if((tofd = socket(AF_INET, SOCK_STREAM, 0)) > -1) {
		Threadlnode->tosocket = tofd;
		ok = TRUE;
	} else
		status = CREATE_SOCKET_ERROR;

	if(ok) {
		/* If we are using transparent proxying kernel support for this port
		 * we must quickly become root again, perform a special bind() and
		 * return to the non-root user and group.
		 */
		if(IPlnode->transproxy) {
			ok = FALSE;

			addr_out = IPlnode->client;
			addr_out.sin_port = 0;

			/* Shortly regain root privileges for the transproxy bind */
			seteuid(0);

			if((bind
				 (tofd, (struct sockaddr *)&addr_out,
				  sizeof(struct sockaddr_in))) > -1)
				ok = TRUE;
			else
				status = TRANSPROXY_BIND_ERROR;

			/* Drop root privileges immediately */
			if((setegid(IPlnode->gid)) || (seteuid(IPlnode->uid))) {
				ok = FALSE;
				status = GIDUID_CHANGE_ERROR;
			}
		}
	}
	
	/* Connect to IP address and port we forward connection to (service) */
	if(ok) {
		if((connect
			 (tofd, (struct sockaddr *)(void *)&Threadlnode->toserver,
			  sizeof(Threadlnode->toserver))) > -1) {

			/* Connected, now forward i/o in both directions */
			struct pollfd polldata[] = {
				{fd, POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL, 0},
				{tofd, POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL, 0}
			};
			timeout = IPlnode->timeout;

			while(ok) {

				/* Wait for fds to change state */
				polldata[0].revents = 0;
				polldata[1].revents = 0;
				selr = poll(polldata, 2, timeout);

				if(selr > 0) {
					/* State changed for at least one descriptor,
					 * find out which and perform appropriate actions
					 */

					/* First detect any error/connection losses */
					for(cnt = 0; cnt < 2; cnt++) {
						selr = polldata[cnt].revents;
						if(selr & POLLERR || selr & POLLHUP || selr & POLLNVAL) {
							ok = FALSE;
							if(cnt == 0)
								status = CLIENT_CLOSED;
							else
								status = SERVICE_CLOSED;
							break;
						}
					}

					if(ok) {

						selr = polldata[0].revents;
						if(selr & POLLIN || selr & POLLPRI) {
							/* Bytes ready to read from client,
							 * forward them to actual service
							 */
							if((len = read(fd, buffer, sizeof(buffer) - 1)) < 1) {
								status = CLIENT_READ_EOF;
								break;
							}
							if(len) {
								bytes_in += len;
								if((write(tofd, buffer, len)) != len) {
									status = CLIENT_WRITE_ERROR;
									break;
								}
							}
						}

						selr = polldata[1].revents;
						if(selr & POLLIN || selr & POLLPRI) {
							/* Bytes available to read from service,
							 * forward them to client
							 */
							if((len = read(tofd, buffer, sizeof(buffer) - 1)) < 1) {
								status = SERVICE_READ_EOF;
								break;
							}
							if(len) {
								if(IPlnode->ftpproxy && IPlnode->transproxy) {
									/* This is an FTP forwarder */
									if((status = ftpproxy
										 (fd, tofd, buffer, sizeof(buffer) - 1,
										  len, &bytes_out, &bytes_in,
										  Threadlnode)) == SERVICE_WRITE_ERROR)
										break;
								} else {
									/* This is a normal forwarder */
									bytes_out += len;
									if((write(fd, buffer, len)) != len) {
										status = SERVICE_WRITE_ERROR;
										break;
									}
								}
							}
						}

					} else
						break;		  /* Not ok ((-: */

				} else {
					status = INPUT_TIMEOUT;
					break;			  /* Timeout */
				}

			}

		} else
			status = SERVICE_CONNECT_ERROR;

	}
	/* Log results */
	time_end = time(NULL);
	time_total = time_end - time_start;
	syslog(LOG_NOTICE,
			 "Closed: ID=0x%04X%08X [%s:%d] - (In: %u, Out: %u, Seconds: %u, Status: %s)",
			 IPlnode->fromport, Threadlnode->ID, ipaddr, IPlnode->fromport,
			 bytes_in, bytes_out, time_total, IPlnode->closed_status[status]);

	return (status);
}


/* This function is used to allow forwarding PASSIVE ftp requests. It also
 * continues forwarding the data connection at the same time
 */
int ftpproxy(int oldfd, int oldtofd, BYTE * buffer, LONG bufsiz, LONG olen,
				 ULONG * obytes_out, ULONG * obytes_in,
				 Threadlistnode * Threadlnode)
{
	IPlistnode *IPlnode = Threadlnode->IPlnode;
	int sock = -1, fd = -1, tofd = -1, selr, type = 0, status = STATUS_UNKNOWN;
	BOOL ok;
	struct sockaddr addr;
	struct sockaddr_in toserver, server, addr_out, sname, *saddr =
		(struct sockaddr_in *)&addr;
	socklen_t addrl;
	int ftpdaddr[4], ftpdport[2], itmp, ftpd_port = 0, our_port;
	LONG len, cnt, dbufferlen = 0;
	BYTE *tmp, *tmp2, dbuffer[1024];

	ULONG bytes_in = 0, bytes_out = 0;
	time_t time_start, time_end, time_total;

	time_start = time(NULL);

	/* Determine passive type */
	ok = TRUE;
	if(!strncmp(buffer, "227", 3))
		type = 227;					  /* PASV */
	else if(!strncmp(buffer, "228", 3))
		type = 228;					  /* LPSV */
	else if(!strncmp(buffer, "229", 3))
		type = 229;					  /* EPSV */
	else
		ok = FALSE;

	if(ok) {
		/* Get the address and port number that ftpd replied with */
		if((tmp = strchr(buffer, '(')) && (tmp2 = strrchr(buffer, ')'))) {
			if(type == 227) {
				/* Normal passive:
				 * 227 Entering Passive Mode (0,0,0,0,192,206)
				 */
				sscanf(tmp, "(%d,%d,%d,%d,%d,%d", &ftpdaddr[0], &ftpdaddr[1],
						 &ftpdaddr[2], &ftpdaddr[3], &ftpdport[0], &ftpdport[1]);
				ftpd_port = (ftpdport[0] << 8) | ftpdport[1];
			} else if(type == 228) {
				/* Long passive:
				 * 228 Entering Long Passive Mode (4,4,0,0,0,0,2,192,207)
				 * In this case: <n> <al> <n>*<al> <pl> <n>*<pl>
				 * As we only support IPv4 at the moment, simply assume 9
				 */
				sscanf(tmp, "(%d,%d,%d,%d,%d,%d,%d,%d,%d", &itmp, &itmp,
						 &ftpdaddr[0], &ftpdaddr[1], &ftpdaddr[2], &ftpdaddr[3],
						 &itmp, &ftpdport[0], &ftpdport[1]);
				ftpd_port = (ftpdport[0] << 8) | ftpdport[1];
			} else if(type == 229) {
				/* Extended passive:
				 * 229 Entering Extended Passive Mode (|||49360|)
				 */
				while(tmp2 > tmp && *tmp2 != '|')
					tmp2--;
				tmp2--;
				while(tmp2 > tmp && *tmp2 != '|')
					tmp2--;
				tmp2++;
				ftpd_port = atoi(tmp2);
			} else
				ok = FALSE;
		} else
			ok = FALSE;
	}

	if(!ok) {
		/* Not a proper passive return code, simply send the data as-is
		 * and continue
		 */
		*obytes_out += olen;
		if(write(oldfd, buffer, olen) != olen)
			return (SERVICE_WRITE_ERROR);
		return (STATUS_UNKNOWN);
	}

	/* Update ftpdaddr[0-3] to represent the current gateway address the client
	 * connects to, our internet IP address
	 */
	len = sizeof(struct sockaddr_in);
	if(!(getsockname(oldfd, (struct sockaddr *)&sname, (socklen_t *) & len))) {
		ftpdaddr[0] = sname.sin_addr.s_addr & 0xff;
		ftpdaddr[1] = (sname.sin_addr.s_addr >> 8) & 0xff;
		ftpdaddr[2] = (sname.sin_addr.s_addr >> 16) & 0xff;
		ftpdaddr[3] = sname.sin_addr.s_addr >> 24;
	} else
		ftpdaddr[0] = ftpdaddr[1] = ftpdaddr[2] = ftpdaddr[3] = 0;

	/* We now know the ftpd server IP and port number to connect to,
	 * as well as the client's address to use to connect to it.
	 * Attempt to connect. If we do successfully, continue. Otherwise
	 * report an error
	 */

	/* Create socket and bind for transproxy */
	ok = FALSE;
	if((tofd = socket(AF_INET, SOCK_STREAM, 0)) > -1) {
		addr_out = IPlnode->client;
		addr_out.sin_port = 0;

		/* Regain root privileges shortly for the transproxy bind */
		seteuid(0);

		if((bind
			 (tofd, (struct sockaddr *)&addr_out,
			  sizeof(struct sockaddr_in))) > -1)
			ok = TRUE;
		else
			status = TRANSPROXY_BIND_ERROR;

		/* Drop root privileges immediately */
		if((setegid(IPlnode->gid)) || (seteuid(IPlnode->uid))) {
			ok = FALSE;
			status = GIDUID_CHANGE_ERROR;
		}
	} else
		status = CREATE_SOCKET_ERROR;

	/* Connect to ftpd port */
	if(ok) {
		ok = FALSE;
		toserver = Threadlnode->toserver;
		toserver.sin_port = htons((UWORD) ftpd_port);
		if((connect
			 (tofd, (struct sockaddr *)(void *)&toserver,
			  sizeof(toserver))) > -1)
			ok = TRUE;
		else
			status = SERVICE_CONNECT_ERROR;
	}

	/* We could connect,  bind a port and listen for connections from the
	 * client's IP only, then send the proper passive reply string telling
	 * the IP which port to connect to.
	 */

	if(ok) {
		status = STATUS_UNKNOWN;
		ok = FALSE;
		if((sock = socket(AF_INET, SOCK_STREAM, 0)) > -1) {
			memset(&server, 0, sizeof(server));
			server.sin_family = AF_INET;
			server.sin_addr.s_addr = sname.sin_addr.s_addr;
			server.sin_port = 0;

			/* Shortly become root to bind the right interface to the port */
			seteuid(0);

			if(bind(sock, (struct sockaddr *)&server, sizeof(server)) >= 0) {
				/* Get the port number we were able to bind */
				len = sizeof(server);
				if(getsockname
					(sock, (struct sockaddr *)&server, (socklen_t *) & len) >= 0) {
					our_port = ntohs(server.sin_port);
					ok = TRUE;
				}
			} else
				status = PASV_BIND_ERROR;
			/* Drop root privileges */
			if((setegid(IPlnode->gid)) || (seteuid(IPlnode->uid))) {
				ok = FALSE;
				status = GIDUID_CHANGE_ERROR;
			}
		}
	} else
		status = CREATE_SOCKET_ERROR;

	if(ok) {
		UBYTE *p = (UBYTE *) & our_port;
		ok = FALSE;

		/* Send passive connection reply to the client */
		if(type == 227) {
			snprintf(buffer, bufsiz - 1,
						"227 Entering Passive Mode (%d,%d,%d,%d,%u,%u)\n",
						ftpdaddr[0], ftpdaddr[1], ftpdaddr[2], ftpdaddr[3], p[1],
						p[0]);
			len = strlen(buffer);
			*obytes_out += len;
			if((write(oldfd, buffer, len)) != len)
				status = CLIENT_WRITE_ERROR;
			else
				ok = TRUE;
		} else if(type == 228) {
			snprintf(buffer, bufsiz - 1,
						"228 Entering Long Passive Mode (4,4,%d,%d,%d,%d,2,%u,%u)\n",
						ftpdaddr[0], ftpdaddr[1], ftpdaddr[2], ftpdaddr[3], p[1],
						p[0]);
			len = strlen(buffer);
			*obytes_out += len;
			if((write(oldfd, buffer, len)) != len)
				status = CLIENT_WRITE_ERROR;
			else
				ok = TRUE;
		} else if(type == 229) {
			snprintf(buffer, bufsiz - 1,
						"229 Entering Extended Passive Mode (|||%d|)\n", our_port);
			len = strlen(buffer);
			*obytes_out += len;
			if((write(oldfd, buffer, len)) != len)
				status = CLIENT_WRITE_ERROR;
			else
				ok = TRUE;
		}
	}

	if(ok) {
		status = INPUT_TIMEOUT;

		if(!(listen(sock, 1))) {
			/* Make sure we get a connection within 60 seconds,
			 * while still forwarding the control port
			 */
			struct pollfd polldata[] = {
				{oldfd, POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL, 0},
				{oldtofd, POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL, 0},
				{sock, POLLIN, 0}
			};

			while(ok) {

				polldata[0].revents = 0;
				polldata[1].revents = 0;
				polldata[2].revents = 0;
				selr = poll(polldata, 3, 60000);
				if(selr > 0) {

					/* First detect any error/connection losses */
					for(cnt = 0; cnt < 2; cnt++) {
						selr = polldata[cnt].revents;
						if(selr & POLLERR || selr & POLLHUP || selr & POLLNVAL) {
							ok = FALSE;
							if(cnt == 0)
								status = CLIENT_CLOSED;
							else
								status = SERVICE_CLOSED;
							break;
						}
					}

					if(ok) {

						selr = polldata[2].revents;
						if(selr & POLLIN) {
							/* We received a connection request */
							addrl = sizeof(struct sockaddr);
							if((fd = accept(sock, &addr, &addrl)) != -1) {
								/* Make sure connection is from actual client's
								 * IP
								 */
								if(saddr->sin_addr.s_addr ==
									IPlnode->client.sin_addr.s_addr)
									break;
								else {
									status = UNAUTHORIZED_ADDRESS;
									ok = FALSE;
									break;
								}
							} else {
								status = PASV_ACCEPT_ERROR;
								ok = FALSE;
								break;
							}
						}

						selr = polldata[0].revents;
						if(selr & POLLIN || selr & POLLPRI) {
							/* Bytes ready to read from client,
							 * forward them to actual service
							 */
							if((len = read(oldfd, buffer, bufsiz - 1)) < 1) {
								status = CLIENT_READ_EOF;
								ok = FALSE;
								break;
							}
							if(len) {
								/* Make sure user can't send those commands
								 * now, delay them if they are sent, appending
								 * them to our delay buffer
								 */
								if((strncasecmp(buffer, "STAT", 4))
									&& (strncasecmp(buffer, "PASV", 4))
									&& (strncasecmp(buffer, "LPSV", 4))
									&& (strncasecmp(buffer, "EPSV", 4))) {
									*obytes_in += len;
									if((write(oldtofd, buffer, len)) != len) {
										status = CLIENT_WRITE_ERROR;
										ok = FALSE;
										break;
									}
								} else {
									/* Append to delayed buffer to be sent once
									 * the transfer is done
									 */
									if(dbufferlen + len < sizeof(dbuffer)) {
										memcpy(dbuffer + dbufferlen, buffer, len);
										dbufferlen += len;
									}
								}
							}
						}

						selr = polldata[1].revents;
						if(selr & POLLIN || selr & POLLPRI) {
							/* Bytes available to read from service,
							 * forward them to client
							 */
							if((len = read(oldtofd, buffer, bufsiz - 1)) < 1) {
								status = SERVICE_READ_EOF;
								ok = FALSE;
								break;
							}
							if(len) {
								*obytes_out += len;
								if((write(oldfd, buffer, len)) != len) {
									status = SERVICE_WRITE_ERROR;
									ok = FALSE;
									break;
								}
							}
						}

					} else
						break;

				} else {
					status = INPUT_TIMEOUT;
					break;
				}

			}
		}
	}

	if(ok) {
		/* We have an actual connection, reply to the client and forward it.
		 * This is not exactly code reuseability, but let's take care of speed
		 * and safety first
		 */

		struct pollfd polldata[] = {
			{fd, POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL, 0},
			{tofd, POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL, 0},
			{oldfd, POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL, 0},
			{oldtofd, POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL, 0}
		};

		status = STATUS_UNKNOWN;

		while(ok) {

			/* Wait for fds to change state with a timeout of 2 minutes */
			polldata[0].revents = 0;
			polldata[1].revents = 0;
			polldata[2].revents = 0;
			polldata[3].revents = 0;
			selr = poll(polldata, 4, 120000);

			if(selr > 0) {
				/* State changed for at least one descriptor,
				 * find out which and perform appropriate actions
				 */

				/* First detect any error/connection losses */
				for(cnt = 0; cnt < 4; cnt++) {
					selr = polldata[cnt].revents;
					if(selr & POLLERR || selr & POLLHUP || selr & POLLNVAL) {
						ok = FALSE;
						if(cnt == 0 || cnt == 2)
							status = CLIENT_CLOSED;
						else
							status = SERVICE_CLOSED;
						break;
					}
				}

				if(ok) {

					selr = polldata[0].revents;
					if(selr & POLLIN || selr & POLLPRI) {
						/* Bytes ready to read from client,
						 * forward them to actual service
						 */
						if((len = read(fd, buffer, bufsiz - 1)) < 1) {
							status = CLIENT_READ_EOF;
							break;
						}
						if(len) {
							bytes_in += len;
							if((write(tofd, buffer, len)) != len) {
								status = CLIENT_WRITE_ERROR;
								break;
							}
						}
					}

					selr = polldata[1].revents;
					if(selr & POLLIN || selr & POLLPRI) {
						/* Bytes available to read from service,
						 * forward them to client
						 */
						if((len = read(tofd, buffer, bufsiz - 1)) < 1) {
							status = SERVICE_READ_EOF;
							break;
						}
						if(len) {
							bytes_out += len;
							if((write(fd, buffer, len)) != len) {
								status = SERVICE_WRITE_ERROR;
								break;
							}
						}
					}

					selr = polldata[2].revents;
					if(selr & POLLIN || selr & POLLPRI) {
						/* Bytes ready to read from client,
						 * forward them to actual service
						 */
						if((len = read(oldfd, buffer, bufsiz - 1)) < 1) {
							status = CLIENT_READ_EOF;
							break;
						}
						if(len) {
							/* Make sure user can't send those commands
							 * immediately, so keep those in our delayed buffer
							 */
							if((strncasecmp(buffer, "STAT", 4))
								&& (strncasecmp(buffer, "PASV", 4))
								&& (strncasecmp(buffer, "LPSV", 4))
								&& (strncasecmp(buffer, "EPSV", 4))) {
								*obytes_in += len;
								if((write(oldtofd, buffer, len)) != len) {
									status = CLIENT_WRITE_ERROR;
									break;
								}
							} else {
								/* Append to delayed buffer to be sent after
								 * the transfer is completed
								 */
								if(dbufferlen + len < sizeof(dbuffer)) {
									memcpy(dbuffer + dbufferlen, buffer, len);
									dbufferlen += len;
								}
							}
						}
					}

					selr = polldata[3].revents;
					if(selr & POLLIN || selr & POLLPRI) {
						/* Bytes available to read from service,
						 * forward them to client
						 */
						if((len = read(oldtofd, buffer, bufsiz - 1)) < 1) {
							status = SERVICE_READ_EOF;
							break;
						}
						if(len) {
							*obytes_out += len;
							if((write(oldfd, buffer, len)) != len) {
								status = SERVICE_WRITE_ERROR;
								break;
							}
						}
					}

				} else
					break;			  /* Not ok ((-: */

			} else {
				status = INPUT_TIMEOUT;
				break;				  /* Timeout */
			}

		}
	}
	/* Now send delayed commands to ftpd control port if any */
	if(dbufferlen) {
		*obytes_in += dbufferlen;
		if((write(oldtofd, dbuffer, dbufferlen)) != dbufferlen)
			status = CLIENT_WRITE_ERROR;
	}
	/* Log results */
	time_end = time(NULL);
	time_total = time_end - time_start;
	syslog(LOG_NOTICE,
			 "Done forwarding passive %d FTP transaction for ID=0x%04X%08X - (In: %u, Out: %u, Seconds: %u, Status: %s)",
			 type, IPlnode->fromport, Threadlnode->ID, bytes_in, bytes_out,
			 time_total, IPlnode->closed_status[status]);

	/* Close our own descriptors */
	if(fd != -1) {
		/* Connection from client to our passive forwarder */
		shutdown(fd, 2);
		close(fd);
	}
	if(sock != -1) {
		/* Our bound socket */
		shutdown(sock, 2);
		close(sock);
	}
	if(tofd != -1) {
		/* Connection with internal ftpd passive port */
		shutdown(tofd, 2);
		close(tofd);
	}

	return (status);
}


/* This function is an identd server which handles a client
 * forwards packets in and out, doing stats and loging. It is both used
 * by phandleclient() and ftpproxy() functions
 */
int handleident(int fd, IPlistnode * IPlnode, Threadlistnode * Threadlnode)
{
	BYTE buffer[1024], ipaddr[20], uid[16];
	BYTE *tmp;
	LONG len, a, b;

	ULONG time_total;
	time_t time_start, time_end;
	int status;


	/* Obtain IP address of client */
	if((tmp = inet_ntoa(IPlnode->client.sin_addr)))
		strncpy(ipaddr, tmp, 19);
	else
		strncpy(ipaddr, "0.0.0.0", 8);

	if(IPlnode->hostname)
		/* Log user's IP and hostname */
		syslog(LOG_NOTICE, "Connect: ID=0x%04X%08X [%s:%d] - (%s) -> mmidentd",
				 IPlnode->fromport, Threadlnode->ID, ipaddr, IPlnode->fromport,
				 IPlnode->hostname);
	else
		/* Log user's IP */
		syslog(LOG_NOTICE, "Connect: ID=0x%04X%08X [%s:%d] -> mmidentd",
				 IPlnode->fromport, Threadlnode->ID, ipaddr, IPlnode->fromport);

	time_start = time(NULL);
	status = STATUS_UNKNOWN;

	if((len = fdgets(buffer, 1000, fd, IPlnode->timeout)) > -1) {

		/* Make sure it's a valid ident request */
		a = 0;
		b = 0;
		/* Parse numbers out of query string */
		a = atol(buffer);
		tmp = buffer;
		while(*tmp && *tmp != ',')
			tmp++;
		if(*tmp == ',') {
			tmp++;
			b = atol(tmp);
		}

		if(a && b) {
			/* Seems a valid request, reply random answer */
			randname(uid, 10);
			snprintf(buffer, 1000, "%ld , %ld : USERID : UNIX : %s\r\n", a, b,
						uid);
			syslog(LOG_NOTICE, "Ident: ID=0x%04X%08X -> [%s]",
					 IPlnode->fromport, Threadlnode->ID, buffer);
			if(fdputs(fd, buffer) != strlen(buffer))
				status = CLIENT_WRITE_ERROR;
			status = IDENT_OK;
		} else {
			/* Invalid request, reply with error */
			snprintf(buffer, 1000, "%ld, %ld : ERROR : INVALID-PORT\r\n", a, b);
			if(fdputs(fd, buffer) != strlen(buffer))
				status = CLIENT_WRITE_ERROR;
			status = BAD_IDENT_REQUEST;
		}

	} else {
		/* Input error */
		if(len == -1)
			status = INPUT_TIMEOUT;
		else if(len == -2)
			status = CLIENT_READ_EOF;
		else
			status = BAD_IDENT_REQUEST;
	}

	/* Log results */
	time_end = time(NULL);
	time_total = time_end - time_start;
	syslog(LOG_NOTICE,
			 "Closed: ID=0x%04X%08X [%s:%d] - (Seconds: %u, Status: %s)",
			 IPlnode->fromport, Threadlnode->ID, ipaddr, IPlnode->fromport,
			 time_total, IPlnode->closed_status[status]);


	return (status);
}


/* Function that generates a random generated name of maximum maxlen chars */
ULONG randname(BYTE * buf, ULONG maxlen)
{
	LONG c;
	ULONG l, i;
	BOOL vowel, lastvowel;
	int fd;

	l = 0;
	if((fd = open("/dev/urandom", O_RDONLY)) != -1) {

		/* Determine lenght of name to be generated (from 3 to maxlen) */
		l = urand32(fd, 3, maxlen);

		/* Now generate random name, taking in consideration dupes and vowels */
		i = urand32(fd, 0, 25);
		i += 'a';
		lastvowel = isvowel(i);
		for(c = 0; c < l; c++) {
			do {
				i = urand32(fd, 0, 25);
				i += 'a';
				vowel = isvowel(i);
			}
			while(lastvowel == vowel);

			lastvowel = vowel;
			buf[c] = (BYTE) i;
		}
		buf[c] = 0;

		close(fd);

	}

	return (l);
}


/* Function that returns a random value of wanted range from fd (urandom) */
ULONG urand32(int fd, ULONG min, ULONG max)
{
	ULONG maxv = 0xFFFFFFFF, val, divd;

	divd = maxv;
	max++;
	divd /= (max - min);

	if((read(fd, &val, 4)) == 4) {
		val /= divd;
		return (min + val);
	}

	return (0);
}


/* Function that returns weither supplied character is a vowel or not */
BOOL isvowel(BYTE c)
{
	LONG i;
	BYTE vowels[] = "aeiouy";

	/* Determine if supplied character is a vowel */
	for(i = 0; i < 6; i++)
		if(vowels[i] == c)
			return (TRUE);

	return (FALSE);
}


/*
 * vim:ts=3:sw=3:cindent:
 */
